import datetime
import os
import csv as csvmod
import time
from typing import Literal
import click
import random
from pdfminer.high_level import extract_text
from talc.utils.cli.io import print_single_case, get_failures_and_warnings
from talc.evals import (
    Document,
    EvalsClient,
    TestCaseWithRubric,
    RunInfo,
    GradedTestCase,
)
from talc.grading import (
    GraderConfig,
    FactualityGraderConfig,
    GradingSet,
    TestCaseResponse,
    ScoringRule,
    GradingPipeline,
)
from talc.utils.grading.defaults import DEFAULT_GRADER_CONFIG


talc_api_key: str | None = os.getenv("TALC_API_KEY", None)
talc_base_url: str | None = os.getenv(
    "TALC_BASE_URL", "http://evals-webserver-webapp.azurewebsites.net"
)


def poll_results(
    client: EvalsClient,
    print_passed: bool,
    out_file: str | None,
    scoring_rules: list[ScoringRule],
):
    while True:
        run_info = client.get_results()
        print(f"Progress: {run_info.completion_progress * 100}%")
        if run_info.completion_progress >= 1:
            print("\n============================================================")
            print("Grading complete.")
            for key in run_info.grades.keys():
                print(f"{key}: {run_info.grades[key] * 100}%")

            print("============================================================\n")

            assert run_info.test_cases is not None

            failures, warnings = get_failures_and_warnings(
                run_info.test_cases, scoring_rules
            )

            passed = [case for case in run_info.test_cases if case not in failures and case not in warnings]

            if out_file is not None:
                with open(out_file, "w", newline="") as csvfile:
                    writer = csvmod.writer(csvfile)
                    writer.writerow(
                        [
                            "id",
                            "question",
                            "expected_answer",
                            "result",
                            "failure",
                            "warning",
                            "grades",
                        ]
                    )
                    for case in run_info.test_cases:
                        writer.writerow(
                            [
                                case.id,
                                case.question,
                                case.expected_response,
                                case.response,
                                True if case in failures else False,
                                True if case in warnings else False,
                                case.grades,
                            ]
                        )

            if print_passed and len(passed) > 0:
                print("=================PASSED=================")
                for case in passed:
                    print_single_case(case, "green")
            if len(warnings) > 0:
                print("=================WARNINGS=================")
                for warning in warnings:
                    print_single_case(warning, "yellow")
            if len(failures) > 0:
                print("=================FAILURES=================")
                for failure in failures:
                    print_single_case(failure, "red")
            if len(failures) > 0:
                return 1
            else:
                return 0

        time.sleep(2)


@click.group()
def evals():
    pass


@evals.command()
@click.option("--count", default=0, help="Number of cases to sample (0 for all).")
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option("--dataset", prompt="Dataset ID", help="The ID of the dataset to load.")
@click.option(
    "--out_file",
    default=None,
    help="Path to CSV file to save results to.",
    required=False,
)
def get_dataset(count: int, dataset: str, api_key: str, out_file: str) -> None:
    """Download or pretty print a test dataset from the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)
    data = client.get_dataset(dataset)

    if count == 0:
        count = len(dataset)

    random.shuffle(data)

    if out_file is not None:
        with open(out_file, "w", newline="") as csvfile:
            writer = csvmod.writer(csvfile)
            writer.writerow(["id", "question", "expected_answer"])
            for case in data[:count]:
                writer.writerow([case.id, case.question, case.expected_response])

    for case in data[:count]:
        print(case.question)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--in_file",
    prompt="Results CSV path",
    help="Path to CSV file to load results from. Columns should be 'id', 'result'. Header row is required and case sensitive.",
)
@click.option("--print_passed", is_flag=True, help="Print the passed test cases.")
@click.option("--out_file", default=None, help="Optionally save output as CSV.")
@click.option(
    "--config",
    default="all_passed",
    help="The name of the grader configuration to use.",
)
def eval(
    api_key: str,
    in_file: str,
    print_passed: bool,
    out_file: str,
    config: str,
):
    """Grade a test run from a CSV. Columns 'id' and 'result' are required, other columns will be ignored. CSV should have a header row."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)

    with open(in_file, newline="") as csvfile:
        reader = csvmod.DictReader(csvfile)

        results = [
            TestCaseResponse(id=row["id"], response=row["result"]) for row in reader
        ]

    if config.endswith(".json"):
        loaded_config = GradingPipeline.parse_file(config)
    else:
        loaded_config = DEFAULT_GRADER_CONFIG

    grading_set = GradingSet(
        responses=results,
        pipeline=loaded_config,  # type: ignore
    )
    print("Grading " + str(len(results)) + " test cases.")
    run_info = client.start_run()
    client.submit_responses(grading_set)
    print("Submitted results. Beginning grading for run ID: " + run_info.id)

    res = poll_results(
        client, print_passed, out_file, loaded_config.scoring_rules
    )
    exit(res)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option("--run_id", prompt="Run ID", help="The ID of the run to get results for.")
@click.option("--print_passed", is_flag=True, help="Print the passed test cases.")
def get_results(api_key: str, run_id: str, print_passed: bool):
    """Get the grades for a test run."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url, run_id)

    res = poll_results(
        client, print_passed, None, DEFAULT_GRADER_CONFIG.scoring_rules
    )

    exit(res)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--csv",
    prompt="CSV path",
    help="Path to CSV file to upload. Column headers should be 'question' and 'expected_answer'. An optional source_text column can be included as well.",
)
@click.option("--name", prompt="Dataset name", help="The name of the dataset.")
def upload_manual_testset(api_key: str, csv: str, name: str):
    """Add a new dataset to the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)

    with open(csv, newline="") as csvfile:
        reader = csvmod.DictReader(csvfile)

        try:
            dataset = [
                TestCaseWithRubric(
                    question=row["question"],
                    expected_response=row["expected_answer"],
                    scenario_data={},
                    id=None,
                    source_content=(
                        [row["source_text"]] if "source_text" in row else []
                    ),
                )
                for row in reader
            ]
        except KeyError:
            raise click.UsageError(
                "CSV file must have columns 'question' and 'expected_answer'. A header is required. An optional source_text column can be included."
            )

    dataset_id = client.upload_dataset(name, dataset)
    print(f"Dataset created with ID: {dataset_id}")


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--in_file",
    prompt="Input CSV path",
    help="Path to CSV file to upload. Columns should be 'question', 'expected_answer' and 'user_answer'. An optional source_text column can be included as well.",
)
@click.option(
    "--name",
    help="The name of the dataset.",
    default="Temp Dataset",
    required=False,
)
@click.option(
    "--out_file",
    help="Name of the output file",
    required=False,
)
def upload_and_grade(api_key: str, in_file: str, name: str, out_file: str | None):
    """Upload a CSV with questions, reference answers, and user answers. Grade the user answers and return the grade. Column names must be 'question', 'expected_answer', and 'user_answer'."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)

    with open(in_file, newline="") as csvfile:
        reader = csvmod.DictReader(csvfile)

        rows = [row for row in reader]

        assert "question" in rows[0]
        assert "expected_answer" in rows[0]
        assert "user_answer" in rows[0]

    dataset = [
        TestCaseWithRubric(
            question=row["question"],
            expected_response=row["expected_answer"],
            scenario_data={},
            id=None,
            source_content=([row["source_text"]] if "source_text" in row else []),
        )
        for row in rows
    ]

    dataset = client.upload_dataset(name, dataset)

    # Download the dataset to get the IDs
    data = client.get_dataset(dataset.id)

    # Map the IDs to the actual responses using the questions as a key

    id_map = {case.question: case.id for case in data}

    results = [
        TestCaseResponse(
            id=id_map[row["question"]],
            response=row["user_answer"] if "user_answer" in row else "",
        )
        for row in rows
    ]

    # Create a run and submit the results
    run_info = client.start_run()

    # MAX TODO: Expose this config to the user
    grading_set = GradingSet(
        responses=results,
        pipeline=DEFAULT_GRADER_CONFIG,
    )
    client.submit_responses(grading_set)
    print("Submitted results. Beginning grading for run ID: " + run_info.id)

    res = poll_results(
        client, True, out_file, DEFAULT_GRADER_CONFIG.scoring_rules
    )

    exit(res)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--in_file",
    prompt="Input folder, single file path, or URL.",
    help="Input folder, single file path, or URL. Multiple allowed.",
    multiple=True,
)
@click.option("--out_file", default=None, help="Path to save the output to.")
@click.option("--dataset_name", default=None, help="The name of the dataset.")
@click.option(
    "--sampling_mode",
    help="The sampling mode to use.",
    type=click.Choice(["main_file", "all_files", "random", "ranked"]),
    default="all_files",
)
@click.option(
    "--generation_mode",
    help="The generation modes to use. Basic types are 'simple', 'complex', 'sweep', and 'hallucination' but others may be supported in custom configurations.",
    default=["sweep"],
    multiple=True,
)
@click.option(
    "--question_limit",
    required=True,
    help="The number of questions to generate. Use 0 to not set a limit.",
)
def generate(
    api_key: str,
    in_file: list[str],
    out_file: str | None,
    dataset_name: str | None,
    sampling_mode: Literal["main_file", "all_files", "random", "ranked"],
    generation_mode: list[
        Literal["simple"]
        | Literal["complex"]
        | Literal["sweep"]
        | Literal["hallucination"]
        | str
    ],
    question_limit: int,
):
    """Generate a dataset from a set of documents."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    if dataset_name is None:
        dataset_name = "questions_" + datetime.datetime.now().strftime(
            "%Y_%m_%d_%H_%M_%S"
        )

    if out_file is None:
        out_file = dataset_name + ".csv"

    client = EvalsClient(api_key, talc_base_url)

    documents: list[Document] = []

    # Figure out what kind of document each file is, and load it into a list of Document objects
    # If it's a URL, send the URL and the server will load it (might want to make this local in the future)
    # If it's a directory, load all the markdown files in the directory
    # If it's a file, load the file
    for f in in_file:
        if f.startswith("http") or f.startswith("https"):
            documents.append(Document(url=f, title=f))
        elif os.path.isdir(f):
            for root, _, files in os.walk(f):
                for path in files:
                    if path.endswith(".md"):
                        with open(os.path.join(root, path), "r") as f:
                            documents.append(Document(content=f.read(), title=f.name))
                    elif path.endswith(".pdf"):
                        documents.append(
                            Document(content=extract_text(os.path.join(root, path)))
                        )
                    else:
                        print(
                            f"Skipping file with unknown extension {path} in directory {root}."
                        )
        elif os.path.isfile(f):
            if f.endswith(".pdf"):
                documents.append(Document(content=extract_text(f), title=f))
            elif f.endswith(".md"):
                with open(f, "r") as f:
                    documents.append(Document(content=f.read(), title=f.name))
            else:
                raise click.UsageError(
                    "Unsupported file type. Only .md and .pdf files are supported."
                )

    dataset_job = client.start_generate_dataset(
        dataset_name,
        documents,
        sampling_mode,
        generation_mode,
        sample_count=question_limit,
    )

    while True:
        dataset_job = client.get_dataset_creation_job(dataset_job.id)
        print("Status: " + dataset_job.status)
        if dataset_job.status == "COMPLETE":
            break
        if dataset_job.status == "ERROR":
            print("Generation failed.")
            exit(1)
        time.sleep(2)

    dataset = client.get_dataset(dataset_job.dataset_id)

    with open(out_file, "w", newline="") as csvfile:
        writer = csvmod.writer(csvfile)
        writer.writerow(["id", "question", "expected_answer"])
        for case in dataset:
            writer.writerow([case.id, case.question, case.expected_response])

    print(f"Dataset created with ID: {dataset_job.dataset_id}")
    print(f"Dataset written to {out_file}")


def main():
    evals()



import os
import csv as csvmod
from textwrap import indent, wrap
from typing import Literal
from termcolor import colored
from talc.evals import (
    Document,
    EvalsClient,
    TestCaseWithRubric,
    RunInfo,
    GradedTestCase,
)
from talc.grading import (
    GraderConfig,
    FactualityGraderConfig,
    GradingSet,
    TestCaseResponse,
    ScoringRule,
    GradingPipeline,
)


# Ugly terminal output functions
def print_single_case(case: GradedTestCase, color: Literal["red", "yellow", "green"]):

    try:
        terminal_width = max(60, os.get_terminal_size().columns)
    except OSError:
        terminal_width = 80

    response_wrapped = "\n".join(
        wrap(f"Response: {case.response}", terminal_width - 10)
    )
    question_wrapped = "\n".join(
        wrap(f"{case.id}: {case.question}\n", terminal_width - 10)
    )

    print(colored(question_wrapped, color, attrs=["bold"]))
    print(indent(response_wrapped, "\t"))
    for grade in case.grades:
        if grade.score >= 1:
            continue
        else:
            grader_color = color
        grade_wrapped = "\n".join(
            wrap(
                f"{grade.grader} {'PASS' if grade.score >= 1.0 else 'FAIL'}: {grade.reason}",
                terminal_width - 10,
            )
        )
        print(colored(grade_wrapped, grader_color))
    print("---------------------------------------------")


def get_failures_and_warnings(
    test_cases: list[GradedTestCase], scoring_rules: list[ScoringRule]
):

    failures = []
    warnings = []

    # Now we need to loop through the rules defined in the scoring rules and
    # determine which cases need to be considered failures vs warnings.
    # This is ugly because it needs to support "any" and "all" scoring.
    for test_case in test_cases:
        for scoring_rule in scoring_rules:
            # determine if the test case has grades for all of the graders for this rule
            eligible = all(
                [
                    any(map(lambda x: x.grader == grader, test_case.grades))
                    for grader in scoring_rule.graders
                ]
            )
            if eligible and scoring_rule.mode == "allow_any_pass":
                all_fail = all(
                    [
                        all(
                            map(
                                lambda x: x.grader != grader or x.score < 1,
                                test_case.grades,
                            )
                        )
                        for grader in scoring_rule.graders
                    ]
                )
                if all_fail:
                    if scoring_rule.level == "fail":
                        failures.append(test_case)
                    elif scoring_rule.level == "warn":
                        warnings.append(test_case)
            elif eligible and scoring_rule.mode == "require_all_pass":
                any_fail = any(
                    [
                        any(
                            map(
                                lambda x: x.grader == grader and x.score < 1,
                                test_case.grades,
                            )
                        )
                        for grader in scoring_rule.graders
                    ]
                )
                if any_fail:
                    if scoring_rule.level == "fail":
                        failures.append(test_case)
                    elif scoring_rule.level == "warn":
                        warnings.append(test_case)

    return failures, warnings



from talc.grading import (
    GradingSet,
    GraderConfig,
    FactualityGraderConfig,
    FewShotConfig,
    FailIfGraderConfig,
    GradingPipeline,
    ScoringRule,
)

DEFAULT_FACTUALITY_GRADER_CONFIG = FactualityGraderConfig(
    grader="FactualityGrader",
    config_name="DefaultFactualityGraderConfig",
    description="The default configuration for the factuality grader.",
    few_shot_examples=[
        FewShotConfig(
            question="What command is used to install the Python SDK for modelscope?",
            correct_answer="The Python SDK for modelscope can be installed using `pip install modelscope` command.",
            user_answer="pip install modelscope",
            grade="PASS",
            reason="The user answer specifies the correct command.",
        ),
    ],
    additional_pass_criteria=[
        "It's ok if the user response includes extra information as long as it does not contradict the reference answer.",
    ],
    additional_fail_criteria=[],
)

DEFAULT_SOURCE_CONTENT_FACTUALITY_GRADER_CONFIG = FactualityGraderConfig(
    grader="SourceContentFactualityGrader",
    config_name="DefaultSourceContentFactualityGraderConfig",
    description="The default configuration for the source content factuality grader.",
    few_shot_examples=[
        FewShotConfig(
            question="What command is used to install the Python SDK for modelscope?",
            correct_answer="The Python SDK for modelscope can be installed using `pip install modelscope` command.",
            user_answer="pip install modelscope",
            grade="PASS",
            reason="The user answer specifies the correct command.",
        ),
    ],
    additional_pass_criteria=[
        "It's ok if the user response includes extra information as long as it does not contradict the source contents.",
    ],
    additional_fail_criteria=[],
)

DEFUALT_FAIL_IF_GRADER_CONFIG = FailIfGraderConfig(
    grader="FailIfGrader",
    config_name="DefaultFailIfGraderConfig",
    description="The default configuration for the fail-if grader.",
    fail_criteria=[
        "The AI responding breaks character by using phrases like 'as an AI agent' or similar.",
    ],
)

DEFAULT_GRADER_CONFIG = GradingPipeline(
    graders=[
        DEFAULT_FACTUALITY_GRADER_CONFIG,
        DEFAULT_SOURCE_CONTENT_FACTUALITY_GRADER_CONFIG,
        DEFUALT_FAIL_IF_GRADER_CONFIG,
    ],
    scoring_rules=[
        ScoringRule(
            graders=[
                "DefaultFactualityGraderConfig",
                "DefaultSourceContentFactualityGraderConfig",
            ],
            level="fail",
            mode="require_all_pass",
        ),
        ScoringRule(
            graders=["DefaultFailIfGraderConfig"],
            level="warn",
            mode="require_all_pass",
        ),
    ],
)



from typing import Any, Literal, Optional
import requests
from pydantic import BaseModel
from talc.grading import GradingSet


class TestCase(BaseModel):
    """Represents a single test case that should be run."""

    id: str
    question: str
    expected_response: str | None


class TestCaseWithRubric(BaseModel):
    """Represents a single test case that should be run."""

    id: str | None
    question: str
    expected_response: str
    scenario_data: dict[str, object]
    source_content: list[str]


class Grade(BaseModel):
    """The output of a single grader on a test case."""

    score: float
    reason: str | None
    grader: str


class GradedTestCase(BaseModel):
    """Represents a single test case that has been graded."""

    id: str
    question: str
    response: str
    expected_response: str
    grading_complete: bool
    grades: list[Grade]


class RunInfo(BaseModel):
    """Represents a test run."""

    id: str
    completion_progress: float
    grades: dict[str, float]
    test_cases: list[GradedTestCase] | None


class Dataset(BaseModel):
    """Represents a dataset."""

    id: str
    friendly_name: str | None


class DatasetCreationJob(BaseModel):
    """Represents the status for a long-running job that creates a dataset in the DB."""

    id: str
    dataset_id: str
    status: str
    error: str | None


class Document(BaseModel):
    """Represents a document used as a ground truth when generating tests."""

    url: Optional[str] = None
    filepath: Optional[str] = None
    title: Optional[str] = None
    content: Optional[str] = None


class APIGenerationArgs(BaseModel):
    """Arguments for creating or getting an API Key."""

    access_token: str


class DatasetCreationArgs(BaseModel):
    """Arguments for creating a dataset."""

    friendly_name: str
    source_docs: list[Document]
    sampling_mode: (
        Literal["main_file"]
        | Literal["all_files"]
        | Literal["random"]
        | Literal["ranked"]
        | None
    )
    generation_modes: list[
        Literal["simple"]
        | Literal["complex"]
        | Literal["sweep"]
        | Literal["hallucination"]
        | str
    ]
    sample_count: int = 10


class EvalsClient:
    """Client for interacting with the Talc evals server."""

    def __init__(self, api_key: str, url: str, run_id: str | None = None) -> None:
        self.url: str = url
        self.run_id: str | None = run_id
        self.api_key: str = api_key

    def start_run(self) -> RunInfo:
        """Create a new test run."""

        response = requests.post(
            f"{self.url}/evals/run/create", headers={"X-TALC-API": self.api_key}
        )
        response.raise_for_status()

        info = RunInfo(**response.json())
        self.run_id = info.id
        return info

    def get_dataset(self, id: str) -> list[TestCase]:
        """Get all test cases in a single dataset from the server."""

        response = requests.get(
            f"{self.url}/evals/dataset/{id}", headers={"X-TALC-API": self.api_key}
        )
        response.raise_for_status()

        return [TestCase(**case) for case in response.json()]

    def submit_responses(self, to_grade: GradingSet) -> None:
        """Submit responses generated by the system for grading."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        data = to_grade.dict()

        response = requests.post(
            f"{self.url}/evals/run/{self.run_id}/submit",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

    def get_results(self) -> RunInfo:
        """Get graded results of a run."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        response = requests.get(
            f"{self.url}/evals/run/{self.run_id}/results",
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return RunInfo(**response.json())

    def upload_dataset(
        self, friendly_name: str, dataset: list[TestCaseWithRubric]
    ) -> Dataset:
        """Upload a new dataset and save it to the database."""

        data = [x.dict() for x in dataset if x.id != "id"]

        response = requests.post(
            f"{self.url}/evals/dataset/create/{friendly_name}",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return Dataset(**response.json())

    def start_generate_dataset(
        self,
        friendly_name: str,
        source_docs: list[Document],
        sampling_mode: (
            Literal["main_file"]
            | Literal["all_files"]
            | Literal["random"]
            | Literal["ranked"]
        ),
        generation_modes: list[
            Literal["simple"]
            | Literal["complex"]
            | Literal["sweep"]
            | Literal["hallucination"]
            | str
        ],
        sample_count: int = 10,
    ) -> DatasetCreationJob:

        data = DatasetCreationArgs(
            friendly_name=friendly_name,
            source_docs=source_docs,
            sampling_mode=sampling_mode,
            generation_modes=generation_modes,
            sample_count=sample_count,
        ).dict()

        response = requests.post(
            f"{self.url}/evals/dataset/generate/",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return DatasetCreationJob(**response.json())

    def get_dataset_creation_job(self, id: str) -> DatasetCreationJob:
        """Get the status of a dataset creation job."""
        response = requests.get(
            f"{self.url}/evals/dataset/generate/{id}",
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return DatasetCreationJob(**response.json())
